#include <windows.h>

#include "autil.h"
#include "aport.h"
#include "aimage.h"

void AImage::rename(wchar_t* sFileName)
{
    if (sFileName)
    {
        if (m_sFileName)
            free(m_sFileName);
        m_sFileName = awcsdup(sFileName);
    }
}

bool AImage::smooth(const AImage &other, const XY & window, int nSparsing)
{
    //TODO: check this and other dimensions
    XY halfWindow = window.half();
    XY size = other.size();
    //TODO: optimization
    for (int x = 0; x < size.m_x; x++)
    {
        for (int y = 0; y < size.m_y; y++)
        {
            XY xy(x,y);
            AColorT<int> col;
            int count = 0;
            for (int dx = -halfWindow.m_x; dx <= halfWindow.m_x; dx += nSparsing)
            {
                for (int dy = -halfWindow.m_y; dy <= halfWindow.m_y; dy += nSparsing)
                {
                    XY point(x+dx,y+dy);
                    if (0 <= point.m_x && point.m_x < size.m_x && 0 <= point.m_y && point.m_y < size.m_y)
                    {
                        col.add(other.at(point));     
                        count++;
                    }
                }
            }
            if (count)
            {
                at(xy) = col.div(count);
            }
        }
    }
    return false;
}

bool AImage::grayScale(const AImage &other, const XY & window, int threshold)
{
    //TODO: check this and other dimensions
    XY halfWindow = window.half();
    XY size = other.size();
    AColor white(255,255,255);
    //TODO: optimization
    for (int x = 0; x < size.m_x; x++)
    {
        for (int y = 0; y < size.m_y; y++)
        {
            XY xy(x,y);
            AColor & this_color = other.at(xy);
            AColorT<int> col;
            int count = 0;
            int maxOfRGB = 0;
            for (int dx = -halfWindow.m_x; dx <= halfWindow.m_x; dx++)
            {
                for (int dy = -halfWindow.m_y; dy <= halfWindow.m_y; dy++)
                {
                    XY point(x+dx,y+dy);
                    if (0 <= point.m_x && point.m_x < size.m_x && 0 <= point.m_y && point.m_y < size.m_y)
                    {
                        AColor & other_color = other.at(point);
                        col.add(other_color);     
                        count++;
                    }
                }
            }
            if (count)
            {
                AColor average_color = col.div(count);
                int distance = average_color.distance(this_color);
                if (distance < threshold)
                {
                    at(xy) = white;
                }
                else
                {
                    int gray = 255 - distance;
                    at(xy) = AColor(gray,gray,gray);
                }
            }
        }
    }
    return true;
}

//TODO: optimization
//based on interval-ed average and max-distance maps
bool AImage::grayScaleFast(const AImage &other, const XY & window, int threshold, bool bNormalizeDistance)
{
    assert(same(other));//check dimensions
    
    XY halfWindow = window.half();
    XY size = other.size();
    
    //prepare the smoothing section
    XY smoothSize((size.m_x + halfWindow.m_x - 1) / halfWindow.m_x, (size.m_y + halfWindow.m_y - 1) / halfWindow.m_y);
    StaticArray2D<AColor> averages(APair<int>(smoothSize.m_x,smoothSize.m_y));
    for (int x = 0; x < smoothSize.m_x; x++)
    {
        for (int y = 0; y < smoothSize.m_y; y++)
        {
            XY xy(x * halfWindow.m_x, y * halfWindow.m_x);
            AColorT<int> col;
            int nCount = 0;
            for (int dx = -halfWindow.m_x; dx <= halfWindow.m_x; dx++)
            {
                for (int dy = -halfWindow.m_y; dy <= halfWindow.m_y; dy++)
                {
                    XY point(xy.m_x + dx,xy.m_y + dy);
                    if (0 <= point.m_x && point.m_x < size.m_x && 0 <= point.m_y && point.m_y < size.m_y)
                    {
                        AColor & other_color = other.at(point);
                        col.add(other_color);     
                        nCount++;
                    }
                }
            }
            assert(nCount > 0);//TODO: exception
            averages.at(x,y) = col.div(nCount);
        }
    }
    
//#define _SMOOTH_AVERAGES_ // this does not make things better really, rather causing "boundary artefacts"
#ifdef _SMOOTH_AVERAGES_
    //smooth the averages again to suppress "boundary effects"
    StaticArray2D<AColor> averages2(APair<int>(smoothSize.m_x,smoothSize.m_y));
    for (int x = 0; x < smoothSize.m_x; x++)
    {
        for (int y = 0; y < smoothSize.m_y; y++)
        {
            AColorT<int> col;
            int nCount = 0;
            for (int dx = -1; dx <= 1; dx++)
            {
                for (int dy = -1; dy <= 1; dy++)
                {
                    XY point(x + dx,y + dy);
                    if (0 <= point.m_x && point.m_x < smoothSize.m_x && 0 <= point.m_y && point.m_y < smoothSize.m_y)
                    {
                        AColor & other_color = averages.at(point.m_x,point.m_y);
                        col.add(other_color);
                        nCount++;
                    }
                }
            }
            assert(nCount > 0);//TODO: exception
            averages2.at(x,y) = col.div(nCount);
        }
    }
    //TODO: don't copy back stuff back but do that in place
    for (int x = 0; x < smoothSize.m_x; x++)
    {
        for (int y = 0; y < smoothSize.m_y; y++)
        {
            averages.at(x,y) = averages2.at(x,y);
        }
    }
#endif    

    StaticArray2D<int> max_distances(APair<int>(smoothSize.m_x,smoothSize.m_y));
    if (bNormalizeDistance)
    {
        for (int x = 0; x < smoothSize.m_x; x++)
        {
            for (int y = 0; y < smoothSize.m_y; y++)
            {
                XY xy(x * halfWindow.m_x, y * halfWindow.m_x);
                AColor & average_color = averages.at(x,y);
                int max_distance = 0;
                for (int dx = -halfWindow.m_x; dx <= halfWindow.m_x; dx++)
                {
                    for (int dy = -halfWindow.m_y; dy <= halfWindow.m_y; dy++)
                    {
                        XY point(xy.m_x + dx,xy.m_y + dy);
                        if (0 <= point.m_x && point.m_x < size.m_x && 0 <= point.m_y && point.m_y < size.m_y)
                        {
                            AColor & other_color = other.at(point);
                            int distance = average_color.distance(other_color);
                            if (max_distance < distance)
                                max_distance = distance;
                        }
                    }
                }
                max_distances.at(x,y) = max_distance;
            }
        }
    }
    
    const AColor white(255,255,255);
    for (int x = 0; x < size.m_x; x++)
    {
        for (int y = 0; y < size.m_y; y++)
        {
            XY this_point(x,y);
            int ax = x/halfWindow.m_x;
            int ay = y/halfWindow.m_y;
            XY average_point(ax,ay);
            AColor & this_color = other.at(this_point);
            AColor & average_color = averages.at(average_point.m_x,average_point.m_y);
            int distance = average_color.distance(this_color);
            if (bNormalizeDistance)
            {
                int max_distance = max_distances.at(ax,ay);
                distance = !max_distance ? 0 : distance * 255 / max_distance;
            }
            if (distance < threshold)
            {
                at(this_point) = white;
            }
            else
            {
                int gray = 255 - distance;
                at(this_point) = AColor(gray,gray,gray);
            }
        }
    }
    return false;
}

bool AImage::grayScaleOnAverage(const AImage &src, const AImage &avg, const XY & normWindow, int threshold, bool bNormalizeDistance, int nSparsing)
{
    assert(same(src) && src.same(avg));//check dimensions
    XY size(src.width(),src.height());
    XY halfWindow = normWindow.half();
    StaticArray2D<int> max_distances(size.m_x,size.m_y);
    if (bNormalizeDistance)
    {
        for (int x = 0; x < size.m_x; x++)
        {
            for (int y = 0; y < size.m_y; y++)
            {
                XY xy(x, y);
                AColor & average_color = avg.at(xy);
                int max_distance = 0;
                for (int dx = -halfWindow.m_x; dx <= halfWindow.m_x; dx += nSparsing)
                {
                    for (int dy = -halfWindow.m_y; dy <= halfWindow.m_y; dy += nSparsing)
                    {
                        XY point(xy.m_x + dx,xy.m_y + dy);
                        if (0 <= point.m_x && point.m_x < size.m_x && 0 <= point.m_y && point.m_y < size.m_y)
                        {
                            AColor & src_color = src.at(point);
                            int distance = average_color.distance(src_color);
                            if (max_distance < distance)
                                max_distance = distance;
                        }
                    }
                }
                max_distances.at(x,y) = max_distance;
            }
        }
    }
    
    const AColor white(255,255,255);
    for (int x = 0; x < size.m_x; x++)
    {
        for (int y = 0; y < size.m_y; y++)
        {
            XY xy(x,y);
            AColor & src_color = src.at(xy);
            AColor & avg_color = avg.at(xy);
            int distance = avg_color.distance(src_color);
            if (bNormalizeDistance)
            {
                int max_distance = max_distances.at(x,y);
                distance = !max_distance ? 0 : distance * 255 / max_distance;
            }
            if (distance < threshold)
            {
                at(xy) = white;
            }
            else
            {
                int gray = 255 - distance;
                at(xy) = AColor(gray,gray,gray);
            }
        }
    }
    return false;
}

bool AImage::bw(const AImage &other, const XY & window)
{
    //TODO: check this and other dimensions
    XY size = other.size();
    XY halfWindow = window.half();
    int nMaxCount = (halfWindow.m_x * 2 + 1) * (halfWindow.m_x * 2 + 1);
    int nCountThreshold = nMaxCount/2;
    AColor black(0,0,0);
    AColor white(255,255,255);
    //TODO: optimization
    for (int x = 0; x < size.m_x; x++)
    {
        for (int y = 0; y < size.m_y; y++)
        {
            XY xy(x,y);
            AColor & this_color = other.at(xy);
            
            int count = 0;
            for (int dx = -halfWindow.m_x; dx <= halfWindow.m_x; dx++)
            {
                for (int dy = -halfWindow.m_y; dy <= halfWindow.m_y; dy++)
                {
                    XY point(x+dx,y+dy);
                    if (0 <= point.m_x && point.m_x < size.m_x && 0 <= point.m_y && point.m_y < size.m_y)
                    {
                        AColor & other_color = other.at(point);
                        if (this_color.equals(other_color))
                            count++;
                    }
                }
            }
            at(xy) = count < nCountThreshold ? black : white;
        }
    }
    return false;
}

//NOTE: assuming grayscale, only first byte of BGR-triad is analysed
bool AImage::createSpots(DynamicArray<ASpot> & lineSpots,const XY & min)
{
    //DynamicArray<ASpot> lineSpots;
    XY size = this->size();
    ASpot spot;
    bool bSpotting = false;

    //first, create linear spots
    int ydir = -1;
#if _HORIZONTAL_STROKING_
    for (int y = ydir > 0 ? 0 : size.m_y - 1; y >= 0 && y < size.m_y; y += ydir)
    {
        for (int x = 0; x < size.m_x; x++)
#else
    for (int x = 0; x < size.m_x; x++)
    {
        for (int y = ydir > 0 ? 0 : size.m_y - 1; y >= 0 && y < size.m_y; y += ydir)
#endif
        {
            XY xy(x,y);
            bool bBlank = at(xy).m_b == 255; //TODO: maybe have a more fair check
            if (bBlank && !bSpotting) //white!
            {
                //keep searching for new spot
                continue;
            }
            else
            if (!bBlank && !bSpotting) //white!
            {
                //mark starting point of new spot
                bSpotting = true;
                spot.m_min.m_x = spot.m_max.m_x = x;
                spot.m_min.m_y = spot.m_max.m_y = y;
            }
            else 
            if (!bBlank && bSpotting)
            {
                //extend confirmed spot area
#if _HORIZONTAL_STROKING_
                spot.m_max.m_x = x;
#else
                spot.m_min.m_y = y;
#endif
            }
            else 
            if (bBlank && bSpotting)
            {
                //complete this spot
#if _HORIZONTAL_STROKING_
                if (x - spot.m_max.m_x > min.m_x)
#else
                if (spot.m_min.m_y - y > min.m_y)
#endif
                {
                    bSpotting = false;
                    lineSpots.add(spot);
                }
            }
        }
        if (bSpotting)
        {
            //complete last spot
            bSpotting = false;
            lineSpots.add(spot);
        }
    }
    return true;
}

bool AImage::drawSpot(const ASpot & spot,const AColor & color,int nBorder)
{
        XY & size = this->size();
        int minx = spot.m_min.m_x;
        int miny = spot.m_min.m_y;
        int maxx = spot.m_max.m_x;
        int maxy = spot.m_max.m_y;
        if (nBorder)
        {
            minx -= nBorder;
            miny -= nBorder;
            maxx += nBorder;
            maxy += nBorder;
        }
        for (int x = max(minx,0); x <= maxx && x < size.m_x; x++)
        {
            if (miny >= 0) 
                at(XY(x,miny)) = color;
            if (miny < maxy && maxy < size.m_y)
                at(XY(x,maxy)) = color;
        }
        for (int y = max(miny + 1,0); y < maxy && y < size.m_y; y++)
        {
            if (minx >=0)
                at(XY(minx,y)) = color;
            if (minx < maxx && maxx < size.m_x)
                at(XY(maxx,y)) = color;
        }
        return true;
}

bool AImage::drawSpots(DynamicArray<ASpot> & spots,const AColor & color,int nBorder)
{
    for (int i = 0; i < spots.size(); i++)
    {
        drawSpot(spots[i],color,nBorder);
    }
    return true;
}


bool ABitmap::load()
{
    m_hImage = LoadImage(
        NULL,
        m_sFileName,
        IMAGE_BITMAP,
        0,
        0,
        LR_LOADFROMFILE|LR_CREATEDIBSECTION
        );
    if (m_hImage)
    {
        GetObject(m_hImage,sizeof(m_bmp),&m_bmp);
        GetObject(m_hImage,sizeof(m_dib),&m_dib);
        return true;
    }
    return false;
}

bool ABitmap::save()
{
    // create file
    FILE * file = awfopen(m_sFileName,L"wb");
    if(!file) 
       return false;

    // save bitmap file headers
    BITMAPFILEHEADER fileHeader;
    fileHeader.bfType      = 0x4d42;
    fileHeader.bfSize      = 0;
    fileHeader.bfReserved1 = 0;
    fileHeader.bfReserved2 = 0;
    fileHeader.bfOffBits   = sizeof(BITMAPFILEHEADER) + sizeof(BITMAPINFOHEADER);

    fwrite((char*)&fileHeader, sizeof(fileHeader),1,file);
    fwrite((char*)&m_dib.dsBmih, sizeof(m_dib.dsBmih),1,file);

    // save dibsection data
    int bytes = (((24*m_bmp.bmWidth + 31) & (~31))/8)*m_bmp.bmHeight;
    fwrite(m_bmp.bmBits, bytes, 1, file);
    fclose(file);
    return true;
}

ABitmap::~ABitmap()
{
    clear();
}

bool ABitmap::create(const ABitmap & other)
{
    clear();//sanity checking to avoid leakage!
    
    //TODO: type checking
    BITMAPINFO info;
    info.bmiHeader = other.m_dib.dsBmih; 
    BYTE* memory = 0;
    m_hImage = CreateDIBSection(0, &info, DIB_RGB_COLORS, (void**)&memory, 0, 0);
    if (m_hImage)
    {
        GetObject(m_hImage,sizeof(m_bmp),&m_bmp);
        GetObject(m_hImage,sizeof(m_dib),&m_dib);
        return true;
    }
    return false;
}

bool ABitmap::capture(int x, int y, int width, int height)
{
    int cx = width  > 0 ? width  : GetSystemMetrics(SM_CXSCREEN);
    int cy = height > 0 ? height : GetSystemMetrics(SM_CYSCREEN);

    BITMAPINFOHEADER infoHeader;
    infoHeader.biSize          = sizeof(infoHeader);
    infoHeader.biWidth         = cx;
    infoHeader.biHeight        = cy;
    infoHeader.biPlanes        = 1;
    infoHeader.biBitCount      = 24;
    infoHeader.biCompression   = BI_RGB;
    infoHeader.biSizeImage     = 0;
    infoHeader.biXPelsPerMeter = 0;
    infoHeader.biYPelsPerMeter = 0;
    infoHeader.biClrUsed       = 0;
    infoHeader.biClrImportant  = 0;

    // dibsection information
    BITMAPINFO info;
    info.bmiHeader = infoHeader; 

    // ------------------
    // THE IMPORTANT CODE
    // ------------------
    // create a dibsection and blit the window contents to the bitmap
    HDC winDC = GetDC(NULL);//Will get he DC for Entire screen 
    HDC memDC = CreateCompatibleDC(winDC);
    BYTE* memory = 0;
    
    HBITMAP bitmap = CreateDIBSection(winDC, &info, DIB_RGB_COLORS, (void**)&memory, 0, 0);
    
    SelectObject(memDC, bitmap);
    BitBlt(memDC, 0, 0, cx, cy, winDC, x, y, SRCCOPY);
    DeleteDC(memDC);
    
    m_hImage = bitmap;
    if (m_hImage)
    {
        GetObject(m_hImage,sizeof(m_bmp),&m_bmp);
        GetObject(m_hImage,sizeof(m_dib),&m_dib);
        return true;
    }
    return false;
}

bool ABitmap::flush(void* pWnd)
{
    HWND hWnd = (HWND)pWnd;
    RECT rect;
    GetWindowRect(hWnd, &rect);
    HDC winDC = GetWindowDC(hWnd);
    HDC memDC = CreateCompatibleDC(winDC);
    SelectObject(memDC, m_hImage);
    BitBlt(winDC, 0, 0, rect.right - rect.left, rect.bottom - rect.top, memDC, 0, 0, SRCCOPY);
    DeleteDC(memDC);
    ReleaseDC(hWnd, winDC);
    return true;
}

bool ABitmap::clear()
{
    if (m_hImage)
    {
        DeleteObject(m_hImage);
        m_hImage = NULL;
    }
    memset(&m_bmp,0,sizeof(m_bmp));
    memset(&m_dib,0,sizeof(m_dib));
    return true;
}


