#ifndef _AIMAGE_H_
#define _AIMAGE_H_

#include "autil.h"
#include "aport.h"
//#ifdef WIN32
//TODO: cutoff win32 specific apart, isolating the virtual bool create(ABitmap & other) = 0;
//TODO: see why it can't be ifdefed
#include <windows.h>
//#endif

class AColor
{
public:
    unsigned char m_r;
    unsigned char m_g;
    unsigned char m_b;
    //unsigned char m_a;
    //AColor(unsigned char b = 0, unsigned char g = 0, unsigned char r = 0, unsigned char a = 0)
    AColor(unsigned char b = 0, unsigned char g = 0, unsigned char r = 0)
    {
        m_r = r;
        m_g = g;
        m_b = b;
        //m_a = a;
    }
    inline bool equals(const AColor & other)
    {
        return m_r == other.m_r && m_g == other.m_g && m_b == other.m_b;
    }
    inline int distance(const AColor & other)
    {
        return (abs((int)m_r - other.m_r) + abs((int)m_g - other.m_g) + abs((int)m_b - other.m_b))/3;
    }
    inline int sum()
    {
        return m_r + m_g + m_b;
    }
};

template <class T> class AColorT
{
public:
    T m_r;
    T m_g;
    T m_b;
    AColorT(T b = 0, T g = 0, T r = 0)
    {
        m_r = r;
        m_g = g;
        m_b = b;
    }
    inline void add(const AColor & other)
    {
        m_r += other.m_r;
        m_g += other.m_g;
        m_b += other.m_b;
    }
    inline AColor div(int denominator)
    {
        AColor ret(
            m_b / denominator,
            m_g / denominator,
            m_r / denominator);
        return ret;
    }
};


class XY
{
public:
    int m_x;
    int m_y;
    XY() : m_x(0), m_y(0) {}
    XY(const XY & other) : m_x(other.m_x), m_y(other.m_y) {}
    XY(int x, int y) : m_x(x), m_y(y) {}
    XY half() const
    {
        return XY(m_x / 2, m_y / 2);
    }
    XY halfPlus() const
    {
        return XY(m_x / 2 + m_x % 2, m_y / 2 + m_y % 2);
    }
    inline bool equals(const XY & other) const
    {
        return m_x == other.m_x && m_y == other.m_y;
    }
    inline int comparex(const XY & other) const
    {
        if (m_x < other.m_x)
            return -1;
        if (m_x > other.m_x)
            return 1;
        //NOTE: y counts from bottom to top
        if (m_y < other.m_y)
            return 1;
        if (m_y > other.m_y)
            return -1;
        return 0;
    }
    inline int comparey(const XY & other) const
    {
        //NOTE: y counts from bottom to top
        if (m_y < other.m_y)
            return 1;
        if (m_y > other.m_y)
            return -1;
        if (m_x < other.m_x)
            return -1;
        if (m_x > other.m_x)
            return 1;
        return 0;
    }
};

class ASpot
{
public:
    XY m_min;
    XY m_max;
    ASpot() {}
    ASpot(int xmin, int ymin, int xmax, int ymax) : m_min(xmin,ymin), m_max(xmax,ymax) {}
    ASpot(const ASpot & other) : m_min(other.m_min), m_max(other.m_max) {}
    ASpot(const XY & min, const XY & max) : m_min(min), m_max(max) { }
    inline bool overlaps(const ASpot & other, const XY & min) const
    {
        return
            (m_min.m_x <= other.m_max.m_x + min.m_x && other.m_min.m_x <= m_max.m_x + min.m_x)
            &&
            (m_min.m_y <= other.m_max.m_y + min.m_y && other.m_min.m_y <= m_max.m_y + min.m_y);
    }
    void merge(const ASpot & other)
    {
        m_min.m_x = min(m_min.m_x,other.m_min.m_x);
        m_min.m_y = min(m_min.m_y,other.m_min.m_y);
        m_max.m_x = max(m_max.m_x,other.m_max.m_x);
        m_max.m_y = max(m_max.m_y,other.m_max.m_y);
    }
    inline int height()
    {
        return m_max.m_y - m_min.m_y + 1;
    }
    inline int width()
    {
        return m_max.m_x - m_min.m_x + 1;
    }
    inline int isPixel()
    {
        return height() == 1 && width() == 1;
    }
    inline bool contains(const XY & point) const
    {
        return m_min.m_x <= point.m_x && point.m_x <= m_max.m_x && m_min.m_y <= point.m_y && point.m_y <= m_max.m_y;
    }
};

class ABitmap;

class AImage
{
protected:
    wchar_t* m_sFileName;
public:
    AImage(wchar_t* sFileName = NULL)
    {
        m_sFileName = NULL;
        rename(sFileName);
    }
    virtual ~AImage()
    {
        if (m_sFileName)
            free(m_sFileName);
    }
    void rename(wchar_t* sFileName);
    virtual bool load() = 0;
    virtual bool save() = 0;
    virtual bool isValid() = 0;
    virtual bool create(const ABitmap & other) = 0;
    virtual bool clear() = 0;
    virtual bool set(const AColor & color) = 0;
    virtual XY size() const = 0;
    virtual int width() const = 0;
    virtual int height() const = 0;
    virtual AColor & at(const XY & xy) const = 0;
    inline bool set(const XY & xy, const AColor & color)
    {
        XY size = this->size();
        if (xy.m_x >=0 && xy.m_x < size.m_x && xy.m_y >=0 && xy.m_y < size.m_y)
        {
            at(xy) = color;
            return true;
        }
        return false;
    }
    inline bool same(const AImage &other) const { return other.width() == width() && other.height() == height(); };
    bool smooth(const AImage &other, const XY & window, int nSparsing = 1);
    bool grayScale(const AImage &other, const XY & window, int threshold);
    bool grayScaleFast(const AImage &other, const XY & window, int threshold, bool bNormalizeDistance = true);
    bool grayScaleOnAverage(const AImage &src, const AImage &avg, const XY & normWindow, int threshold, bool bNormalizeDistance, int nSparsing = 1);
    bool bw(const AImage &other, const XY & window);
    bool createSpots(DynamicArray<ASpot> & spots,const XY & min);
    bool inline drawSpot(const ASpot & spot,const AColor & color,int nBorder);
    bool drawSpots(DynamicArray<ASpot> & spots,const AColor & color,int nBorder = 0);
};

class ABitmap : public AImage
{
private:
    HANDLE m_hImage;
    BITMAP m_bmp;
    DIBSECTION m_dib;
public:
    ABitmap(wchar_t* sFileName = NULL) : AImage(sFileName) {}
    virtual bool isValid()
    {
        return m_hImage && m_bmp.bmWidth > 0 && m_bmp.bmHeight > 0;
    }
    virtual int width() const 
    {
        return m_bmp.bmWidth;
    }
    virtual int height() const 
    {
        return m_bmp.bmHeight;
    }
    virtual XY size() const 
    {
        return XY(m_bmp.bmWidth, m_bmp.bmHeight);
    }
    virtual bool load();
    virtual bool save();
    virtual ~ABitmap();
    void save(wchar_t* sFileName)
    {
        rename(sFileName);
        save();
    }
    virtual bool create(const ABitmap & other);
    virtual bool clear();
    virtual bool set(const AColor &color)
    {
        int bytes = (((24*m_bmp.bmWidth + 31) & (~31))/8)*m_bmp.bmHeight;
        for (int i = 0; i < bytes; i+=3)
        {
            unsigned char * pByte = (unsigned char *) m_bmp.bmBits;
            pByte[i + 0] = color.m_r;
            pByte[i + 1] = color.m_g;
            pByte[i + 2] = color.m_b;
        }
        return true;
    }   
    virtual AColor & at(const XY & xy) const
    {
        unsigned char * pByte = (unsigned char *) m_bmp.bmBits;
        int nRowBytes = (((24*m_bmp.bmWidth + 31) & (~31))/8);
        pByte += nRowBytes * xy.m_y + xy.m_x * 3;
        return *(reinterpret_cast<AColor *>(pByte));
    }
    //Specific operations
    bool capture(int x = 0, int y = 0, int width = 0, int height = 0);//capture entire Win32 desktop
    bool flush(void* pWnd);//flush contents to Win32 window
};

#endif //_AIMAGE_H_
