/*
 * Copyright 1997-1999 by Anton Kolonin,
 * All rights reserved.
 *
 * This software is the proprietary information of Anton Kolonin. 
 * You shall not disclose this and shall not use it in any way 
 * without of written permission of Anton Kolonin.
 */
package com.vm.visual;


import java.awt.Event;
import java.awt.Point;
import java.awt.Graphics;
import java.util.Vector;

import com.vm.app.VMResource;
import com.vm.mind.MIAnything;
import com.vm.mind.MIThing;
import com.vm.visual.VMNode;
import com.vm.visual.objects.VM3DObject;
import com.vm.visual.objects.VM3DReference;

public class VMCameraProjector extends VMProjector
{
	VMWanderer m_wan = null;//20061111
	
    public VMCameraProjector(VMSpace i_space, VMPanel i_panel)
    {
        super(i_space,i_panel);
        m_locatorSize = Integer.valueOf(i_space.vmw_container.getParameter("locatorSize","10")).intValue();
        
        m_wan = new VMWanderer(i_panel);//20061111
    }

    /** indicates that objects has been changed (created, deleted, moved, or amended internaly) 
     * so new child objects should be created from parents */
    private boolean m_bDataChanged = true;

    /** indicates that view has been changed (created, deleted, moved, or amended internaly) 
     * so that existing objects should get resorted */
    private boolean m_bViewChanged = true;
    
    //int cameraX, cameraY, cameraZ;
    
    /** max visibility distance */
    int maxVisibility = 2000000;
    
    int minVisibility = 1;
    
    /** between camera and view plane */
    //int distance = 1000;
    int distance = 1000;
    //int distance = 10;

    int increment = 10;

    double m_theta=space.m_viewpoint.m_theta;
    double m_phi=space.m_viewpoint.m_phi;
    double cos_theta;
    double sin_theta;
    double cos_phi;
    double sin_phi;
    double k1;
    double k3;
    double k4;
    double k5;
    double k6;
    double k7;
    double k8;
    double k9; 
    
    public VMNode light_node;

    private transient float[] m_middlePoint = new float[3];

    
    //ArrayList references = new ArrayList(); // JDK 1.2
    Vector references = new Vector(); 

    /** internal locator size used to locate the objects */
    private int m_locatorSize;
    
    static final int X = 0;
    static final int Y = 1;
    static final int Z = 2;
    
    /**
        @return null if not available
        */
    public Point getPoint(float[] point)
    {
//System.out.println("  settings: t="+space.m_viewpoint.m_theta+" p="+space.m_viewpoint.m_phi+" x="+centerX+" y="+centerY);

        transform3D2D(point);
        
        if (!valid(point))
            return null;
        else    
            return new Point(Math.round(point[X]),Math.round(point[Y]));
    }
        
    /**
        @return null if not available
        */
    public Point getPoint(int x, int y, int z)
    {
        return getPoint(new float[]{x,y,z});
    }   
    
    public boolean valid(float[] p)
    {
        return (minVisibility<p[2] && p[2]<maxVisibility && Math.abs(p[0])<maxVisibility && Math.abs(p[1])<maxVisibility);
    }
    

    /**
        // rotation around Y
        float x  = (float)( point[X]*cos_theta + point[Z]*sin_theta);
        point[Z] = (float)(-point[X]*sin_theta + point[Z]*cos_theta);
        point[X] = x;

        // rotation around X
        float y  = (float)( point[Y]*cos_phi - point[Z]*sin_phi);
        point[Z] = (float)( point[Y]*sin_phi + point[Z]*cos_phi);
        point[Y] = y;
        */        
    public void transform3D(float[] point)
    {
        point[X] -= space.m_viewpoint.m_x;
        point[Y] -= space.m_viewpoint.m_y;
        point[Z] -= space.m_viewpoint.m_z;

        // rotation around Y
        // +
        // rotation around X
        float x  = (float)( point[X]*cos_theta + point[Z]*sin_theta);
        float y  = (float)( point[Y]*cos_phi + point[X]*sin_theta*sin_phi - point[Z]*cos_theta*sin_phi);
        float z  = (float)( point[Y]*sin_phi - point[X]*sin_theta*cos_phi + point[Z]*cos_theta*cos_phi);
        
        point[X] = x;
        point[Y] = y;
        point[Z] = z;
    }
  
    public void project3D2D(float[] point)
    {
//        double k = distance / Math.sqrt(point[Z]*point[Z]+point[X]*point[X]+point[Y]*point[Y]);  //?
        double k = distance / point[Z];  
        point[X]=(float)(panel.centerX+point[X]*k);
        point[Y]=(float)(panel.centerY+point[Y]*k);
    }
        
    /**
     * 
     * @param point
     * @return
     * @since 20060529
     */
    public int[] project3D2Dint(float[] point)
    {
    	int[] res = new int[2];
    	project3D2D(point);
        res[0]=(int)Math.round(point[X]);
        res[1]=(int)Math.round(point[Y]);
    	return res;
    }

    // good !!!
    public void transform3D2D(float[] point)
    {
        point[X] -= space.m_viewpoint.m_x;
        point[Y] -= space.m_viewpoint.m_y;
        point[Z] -= space.m_viewpoint.m_z;

        // rotation around Y
        // +
        // rotation around X
        /*
        double _x = (point[X]*cos_theta + point[Z]*sin_theta);
        double _y = (point[Y]*cos_phi + point[X]*sin_theta*sin_phi - point[Z]*cos_theta*sin_phi);
        double z  = (point[Y]*sin_phi - point[X]*sin_theta*cos_phi + point[Z]*cos_theta*cos_phi);
        double x = panel.centerX+_x*distance/z;
        double y = panel.centerY+_y*distance/z;
        */
        /*
        double z = point[Y]*sin_phi - point[X]*sin_theta*cos_phi + point[Z]*cos_theta*cos_phi;
        double x = panel.centerX+(point[X]*cos_theta + point[Z]*sin_theta)*distance/z;
        double y = panel.centerY+(point[Y]*cos_phi + point[X]*sin_theta*sin_phi - point[Z]*cos_theta*sin_phi)*distance/z;
        */
        
        double k1;
        double k3;
        double k4;
        double k5;
        double k6;
        double k7;
        double k8;
        double k9;
        
        k1=cos_theta*distance;
        k3=sin_theta*distance;
        k4=sin_theta*sin_phi*distance;
        k5=cos_phi*distance;
        k6=-cos_theta*sin_phi*distance;
        k7=-sin_theta*cos_phi;
        k8=sin_phi;
        k9=cos_theta*cos_phi;
        
        double z = point[X]*k7 + point[Y]*k8 + point[Z]*k9;
        double x = panel.centerX+(point[X]*k1 + point[Z]*k3)/z;
        double y = panel.centerY+(point[X]*k4 + point[Y]*k5 + point[Z]*k6)/z;
        
        point[X] = (float) x;
        point[Y] = (float) y;
        point[Z] = (float) z;
    }


    /*
    Translates screen x,y into 3D point (Y-fixed)
    
    point[Z]*(dx*k9 - k3)=point[X]*k1 - point[X]*dx*k7 - point[Y]*dx*k8;
    double c2=(dx*k9 - k3);
    point[Z]=point[X]*k1/c2 - point[X]*dx*k7/c2 - point[Y]*dx*k8/c2;
    point[X] = ((point[Y]*k5 - point[Y]*dx*k8*k6/c2 - point[Y]*dy*k8 + point[Y]*dx*k8*dy*k9/c2)/(dy*k7 + k1*dy*k9/c2 - dx*k7*dy*k9/c2 + dx*k7*k6/c2 - k4 - k1*k6/c2)); 
    */
    void inverseXZ(int x, int y, float[] point)
    {
        point[X] -= space.m_viewpoint.m_x;
        point[Y] -= space.m_viewpoint.m_y;
        point[Z] -= space.m_viewpoint.m_z;
        int dx=x-panel.centerX;
        int dy=y-panel.centerY;
        double c2=(dx*k9 - k3);
        point[X] = (float) ((point[Y]*k5 - point[Y]*dx*k8*k6/c2 - point[Y]*dy*k8 + point[Y]*dx*k8*dy*k9/c2)/(dy*k7 + k1*dy*k9/c2 - dx*k7*dy*k9/c2 + dx*k7*k6/c2 - k4 - k1*k6/c2)); 
        point[Z] = (float) (point[X]*k1/c2 - point[X]*dx*k7/c2 - point[Y]*dx*k8/c2);
        point[X] += space.m_viewpoint.m_x;
        point[Y] += space.m_viewpoint.m_y;
        point[Z] += space.m_viewpoint.m_z;
    }


    /*  
    Translates screen x,y into 3D point (Z-fixed)
    
    z=point[X]*k7 + point[Y]*k8 + point[Z]*k9;
    x-panel.centerX=(point[X]*k1 + point[Z]*k3)/z;
    y-panel.centerY=(point[X]*k4 + point[Y]*k5 + point[Z]*k6)/z;

    dx=x-panel.centerX;
    dy=y-panel.centerY

    z=point[X]*k7 + point[Y]*k8 + point[Z]*k9;
    dx*z=(point[X]*k1 + point[Z]*k3);
    dy*z=(point[X]*k4 + point[Y]*k5 + point[Z]*k6);

    point[X]*dx*k7 + point[Y]*dx*k8 + point[Z]*dx*k9=point[X]*k1 + point[Z]*k3;
    point[X]*dy*k7 + point[Y]*dy*k8 + point[Z]*dy*k9=point[X]*k4 + point[Y]*k5 + point[Z]*k6;
    
    c1=(dx*k7 - k1);
    point[X] = point[Z]*k3/c1 - point[Y]*dx*k8/c1 - point[Z]*dx*k9/c1;
    
    point[Y]*(dy*k8 - k5 - dx*k8*dy*k7/c1 + dx*k8*k4/c1) =point[Z]*k3*k4/c1  - point[Z]*dx*k9*k4/c1  + point[Z]*k6 + point[Z]*dx*k9*dy*k7/c1 - point[Z]*k3*dy*k7/c1 - point[Z]*dy*k9;

    point[Y] = ((point[Z]*k3*k4/c1  - point[Z]*dx*k9*k4/c1  + point[Z]*k6 + point[Z]*dx*k9*dy*k7/c1 - point[Z]*k3*dy*k7/c1 - point[Z]*dy*k9)/(dy*k8 - k5 - dx*k8*dy*k7/c1 + dx*k8*k4/c1));
    */
    void inverseXY(int x, int y, float[] point)
    {
        point[X] -= space.m_viewpoint.m_x;
        point[Y] -= space.m_viewpoint.m_y;
        point[Z] -= space.m_viewpoint.m_z;

        int dx=x-panel.centerX;
        int dy=y-panel.centerY;
        double c1=(dx*k7 - k1);

        point[Y] = (float) ((point[Z]*k3*k4/c1  - point[Z]*dx*k9*k4/c1  + point[Z]*k6 + point[Z]*dx*k9*dy*k7/c1 - point[Z]*k3*dy*k7/c1 - point[Z]*dy*k9)/(dy*k8 - k5 - dx*k8*dy*k7/c1 + dx*k8*k4/c1));
        point[X] = (float) (point[Z]*k3/c1 - point[Y]*dx*k8/c1 - point[Z]*dx*k9/c1);

        point[X] += space.m_viewpoint.m_x;
        point[Y] += space.m_viewpoint.m_y;
        point[Z] += space.m_viewpoint.m_z;
    }

    public boolean getViewChanged()//20061121
    {
		return m_bViewChanged;
    }  
    
    private synchronized void setViewChanged(boolean changed)//20061121
    {
		m_bViewChanged = changed;
    }  

    void init()
    {
    	setViewChanged(true);//20061121
    	
    	/*
    	//not sure if we need this, so disable //20061111 
        if (space.control==null)
        {
            space.control=new VMControl(panel,space.m_viewpoint.m_theta,space.m_viewpoint.m_phi);
        }
        else
        {
            space.control.getViewSettings();
        }
        */
        
        if (m_wan!=null)//20061121
        {
        	//always try to reset the wanderer though it may just stay in the same state
        	//this may be redundant, so reconsider this later!!!
        	m_wan.toggle(space.getAnimateMode());//20061202
        	if ((space.getAimMode() && m_wan.wander(this)) || space.getAnimateMode())//20061202
        	{
        		// redundant, not good //20061111
				if (space.control!=null)//20061111
					space.control.getViewSettings();
				setScrollbars(0,0);
        	}
        }

        m_theta=space.m_viewpoint.m_theta;
        m_phi=space.m_viewpoint.m_phi;
        
        //angles in radians
        float factor=(float)(Math.atan(1.0F)/45.0F);
        m_theta*=factor;m_phi*=factor;

        cos_theta=Math.cos(m_theta);
        sin_theta=Math.sin(m_theta);
        cos_phi=Math.cos(m_phi);
        sin_phi=Math.sin(m_phi);
        
        k1=cos_theta*distance;
        k3=sin_theta*distance;
        k4=sin_theta*sin_phi*distance;
        k5=cos_phi*distance;
        k6=-cos_theta*sin_phi*distance;
        k7=-sin_theta*cos_phi;
        k8=sin_phi;
        k9=cos_theta*cos_phi;
    }

    void reflectViewChange()//20061111
    {
        if (space.control!=null)//20061111
			space.control.getViewSettings();
        setScrollbars(0,0);
        init();
    }
    
    void setScrollbars(int xmax,int ymax)
    {
        boolean tsb = trueScrollBars();
        if (tsb)
        {
            panel.horz.setValues((int)-space.m_viewpoint.m_theta,10,-180,190);//20061111
            panel.vert.setValues((int)space.m_viewpoint.m_phi  ,10,-180,190);//20061111
        }
        else
        {
            panel.horz.setValues((int)-space.m_viewpoint.m_theta,0,-180,180);//20061111
            panel.vert.setValues((int)space.m_viewpoint.m_phi  ,0,-180,180);//20061111
        }
    }
    
    
    /** 
        prepaint callback - never scroll, change angles 
        */
    Point getScrollOffset()
    {
        int theta = -panel.horz.getValue();
        int phi   = panel.vert.getValue();
        if (theta!=space.m_viewpoint.m_theta || phi!=space.m_viewpoint.m_phi)
        {
            space.m_viewpoint.m_theta=theta;
            space.m_viewpoint.m_phi=phi;

            if (space.control!=null)//20061111
            	space.control.getViewSettings();
            
            init();
        }
        panel.setCenter(panel.size().width/2,panel.size().height/2);
        return new Point(0,0);
    }

    /** never paint 2D shadows in 3D */
    void paintShadows(Graphics offgraphics)
    {
    }

    void updateReferences(Vector refs)//20061111
    {
    	if (refs!=null)
    	{
    		int length = refs.size();
    		for (int i=0;i<length;i++)
    		{
    			VM3DReference ref = (VM3DReference)refs.elementAt(i); 
                float p[]=ref.object.center3D();
                if (p!=null)
                {
                    transform3D(p);
                    ref.depth = (int)p[2];
                }
    		}
    	}
    }
    
    void addReferences(VM3DObject o)
    {
        VM3DObject c[] = o.children();
        if (c==null)
        {
                float p[]=o.center3D();
                if (p == null){//sticky depthless topmost object//20060518
                  references.addElement(new VM3DReference(o,Integer.MIN_VALUE));
                }else{
                  transform3D(p);
                  if (valid(p))
                  {
                     references.addElement(new VM3DReference(o,(int)p[2]));
                  }
                  else
                  {
//System.out.println("        ignored "+o.getClass().getName()+" at "+p[0]+" "+p[1]+" "+p[2]);
                  }
                }
        }
        else
        {
            for (int i=0;i<c.length;i++)
            {
            	//skip invalid null ojects - may be returned by VMSurface
            	//in place of undefined grid cells
            	if (c[i]!=null)
            		addReferences(c[i]);
            }
        }
    }

    /*
     * used by Wanderer
     */
    public float[] middlePoint()//20061111
    {
    	//if (m_bDataChanged)//we can not track this really!
    	{
    		int cnt = 0;
    		m_middlePoint[0]=m_middlePoint[1]=m_middlePoint[2]=0;
            for (int i=space.m_viewpoint.m_nodes.getMin();
            	i<=space.m_viewpoint.m_nodes.getMax();
            	i=space.m_viewpoint.m_nodes.getNext(i))
            {
            	VM3DObject o = (VM3DObject)space.m_viewpoint.m_nodes.get(i); 
            	float[] p = o.center3D();
            	int weight = o.getWeight(); 
            	if (p!=null && weight>0)
            	{
            		cnt += weight;
            		m_middlePoint[0]+=p[0]*weight;
            		m_middlePoint[1]+=p[1]*weight;
            		m_middlePoint[2]+=p[2]*weight;
            	}
            }
            if (cnt>0)
            {
        		m_middlePoint[0]/=cnt;
        		m_middlePoint[1]/=cnt;
        		m_middlePoint[2]/=cnt;
            }
    	}
    	return m_middlePoint;
    }
    
    void paintObjects(Graphics offgraphics)
    {
            MIThing t = panel.space.m_world.thing("sun");
            light_node = (t==null)? null: panel.space.m_viewpoint.getNode(t);

            // recreate references to complex objects 
            if  (m_bDataChanged) {//20060605
	            // clear 3D references
	            references.removeAllElements();
	            // reference objects in 3D
	            for (   int i=space.m_viewpoint.m_nodes.getMin();
	                    i<=space.m_viewpoint.m_nodes.getMax();
	                    i=space.m_viewpoint.m_nodes.getNext(i))
	            {
	                 addReferences((VM3DObject)space.m_viewpoint.m_nodes.get(i));
	            }
            }
            else//20061111
            if (m_bViewChanged) // need to change depths because viewpoint is changed
            	updateReferences(references);
            
            if (m_bViewChanged || m_bDataChanged)//20060605
            	VMToolkit.sort(references);

            offgraphics.setColor(panel.arrowColor);
            
            for (int i=0;i<references.size();i++)    
            {
                Object o = references.elementAt(i);
                ((VM3DReference)o).object.paintAttributes(panel,offgraphics);
            }

            for (int j=0;j<references.size();j++)    
            {
                Object o = references.elementAt(j);
                try {
                    ((VM3DReference)o).object.paintObject(panel,offgraphics);
                } catch (Exception e){
                    panel.space.message("Camera.paint: "+e);        
                }
            }

            // cleanup change notifications //20061121
            m_bDataChanged = false;//20060605
            setViewChanged(false);//20060605//20061121
    }

    synchronized protected VM3DObject locateNode(int x, int y)
    {
        double bestdist2 = Double.MAX_VALUE;
        double bestdepth = Double.MAX_VALUE; 
        VM3DObject n;
        VM3DReference r;
        Point p;
        double depth;
        VM3DObject located=null;
        
        //MIList list = (references!=null)? references: space.m_viewpoint.m_nodes;
        /*20061020
        for (   int i=space.m_viewpoint.m_nodes.getMin();
                i<=space.m_viewpoint.m_nodes.getMax();
                i=space.m_viewpoint.m_nodes.getNext(i))
        {
            n=(VM3DObject)space.m_viewpoint.m_nodes.get(i);
            */
        int max = (references!=null)? references.size(): 0;
        for (int i=0;i<max;i++)
        {
        	r=(VM3DReference)references.elementAt(i);
        	n = r.object;
            float[] g = n.center3D();
            if (g!=null)
            {
                p = getPoint(g);
            	depth = r.depth;
            }
            else
            {
            	p = n.getPoint();
            	depth = 0;
            }
            if (p!=null)
            {
                int dx = x-p.x;
                int dy = y-p.y;
                double dist2 = dx*dx + dy*dy;
                if (dist2<bestdist2 && depth<=bestdepth  
	            	&& Math.abs(dx)<m_locatorSize && Math.abs(dy)<m_locatorSize)
                {
                    located = n;
                    bestdist2 = dist2;
                    bestdepth = depth;
                }
            }//(p!=null)
        }
        if (located!=null)
        {
        	return located.getParent();
        }    
        return null;
    }
    
    //20061022
    public void reflectChange()
    {
        m_bDataChanged = true;
    }

    public void reflectDrag(Event evt,VMNode node, int x, int y)
    {
        m_bDataChanged = true;//20060605
        
        float[] p = new float[] {node.m_x,node.m_y,node.m_z};
        if ((evt.modifiers&Event.SHIFT_MASK)!=0)
            inverseXZ(x,y,p);
        else
            inverseXY(x,y,p);
            
        node.m_x=Math.round(p[X]);
        node.m_y=Math.round(p[Y]);
        node.m_z=Math.round(p[Z]);

        space.m_world.notify(space.m_viewpoint,node);
    }    

    public void reflectDelete()
    {
    	m_bDataChanged = true;//20060605
        super.reflectDelete();
    }

    void reflectInsert(MIAnything someThing)//20060605
    {
    	m_bDataChanged = true;//20060605
        super.reflectInsert(someThing);
    }
    
    /*
    HOME    -look straight
    END     -turn around
    UP      -forward
    DOWN    -backward
    LEFT    -rotate left
    RIGHT   -rotate right
    PGUP    -look up
    PGDN    -look down
    SHIFT+UP    -run forward 
    SHIFT+DOWN  -run backward
    SHIFT+LEFT  -move left
    SHIFT+RIGHT -move right
    SHIFT+PGUP  -fly up
    SHIFT+PGDN  -fly down
    
    MOUSE       - moves nodes within vertical plane 
    SHIFT+MOUSE - moves nodes within horizontal plane 
    
    */
    public boolean action(Event evt, Object arg)
    {
        if (evt.target==space.world_url)
            return false;
    	if (arg instanceof float[])//20061121
    	{
    		float[] msg = (float[])arg; 
    	    space.m_viewpoint.m_x = msg[0]; 
    	    space.m_viewpoint.m_y = msg[1]; 
    	    space.m_viewpoint.m_z = msg[2]; 
    	    space.m_viewpoint.m_theta = msg[3]; 
    	    space.m_viewpoint.m_phi = msg[4]; 
			reflectViewChange();
            return true;
    	}
    	else
        if ((evt.modifiers&Event.SHIFT_MASK)==0)
        {
            //no shift key
            if (evt.key==Event.HOME)
            {
                space.m_viewpoint.m_phi=0;
                reflectViewChange();//20061111
                return true;
            }
            else
            if (evt.key==Event.END)
            {
                space.m_viewpoint.m_phi=0;
                space.m_viewpoint.m_theta+=180;
                if (space.m_viewpoint.m_theta>180)
                    space.m_viewpoint.m_theta-=360;
				reflectViewChange();//20061111
                return true;
            }
            else
            if (evt.key==Event.PGUP)
            {
            	//move straight up
                space.m_viewpoint.m_y-=increment;
                init();
                return true;
            }
            else
            if (evt.key==Event.PGDN)
            {
            	//move straight down
                space.m_viewpoint.m_y+=increment;
                init();
                return true;
            }
            else
            if (evt.key==Event.UP)
            {
            	//move forward
                space.m_viewpoint.m_z+=increment*cos_theta;
                space.m_viewpoint.m_x-=increment*sin_theta;
                init();
                return true;
            }
            else
            if (evt.key==Event.DOWN)
            {
            	//move backward
                space.m_viewpoint.m_z-=increment*cos_theta;
                space.m_viewpoint.m_x+=increment*sin_theta;
                init();
                return true;
            }
            else
            if (evt.key==Event.RIGHT)
            {
            	//move right
                space.m_viewpoint.m_z+=increment*sin_theta;
                space.m_viewpoint.m_x+=increment*cos_theta;
                init();
                return true;
            }
            else
            if (evt.key==Event.LEFT)
            {
            	//move left
                space.m_viewpoint.m_z-=increment*sin_theta;
                space.m_viewpoint.m_x-=increment*cos_theta;
                init();
                return true;
            }
        }    
        else 
        {
            // shift key  
            if (evt.key==Event.UP)
            {
            	// raise head 
            	space.m_viewpoint.m_phi--;
    			reflectViewChange();//20061111
                return true;
            }
            else
            if (evt.key==Event.DOWN)
            {
            	// lower head
            	space.m_viewpoint.m_phi++;
    			reflectViewChange();//20061111
                return true;
            }
            if (evt.key==Event.RIGHT)
            {
            	//turn head right
                space.m_viewpoint.m_theta--;
                if (space.m_viewpoint.m_theta<-180)
                    space.m_viewpoint.m_theta+=360;
				reflectViewChange();//20061111
                return true;
            }
            else
            if (evt.key==Event.LEFT)
            {
            	//turn head left
                space.m_viewpoint.m_theta++;
                if (space.m_viewpoint.m_theta>180)
                    space.m_viewpoint.m_theta-=360;
				reflectViewChange();//20061111
                return true;
            }
        }
        return false;
    }
    

    public String[] clickActions()
    {
//        return new String[]{"Edit","Dump","Goto","Delete","Focus","Clone","Ungroup"};
        return new String[]{
        		VMResource.getString("Edit"),
        		VMResource.getString("Dump"),
        		VMResource.getString("Goto"),
        		VMResource.getString("Delete")
        		//VMResource.getString("Clone")//commented out for Space Work only!!!!!//20061023
        		};
    }
    
    public String[] dragActions()
    {
//        return new String[]{"Move","Link","Size","Group"};
        return new String[]{"Move","Link"};
    }
    
    public boolean clickAction(String action, int x, int y)
    {
        if (action.equals("Focus"))
        {
            VM3DObject o = locateNode(x,y);
            VMNode n = (o instanceof VMNode)? (VMNode)o: null; 
            if (n!=null)
            {
                space.m_viewpoint.m_x=n.m_x;
                space.m_viewpoint.m_y=n.m_y;
                space.m_viewpoint.m_z=n.m_z;
                init();
//System.out.println("Focused!");
                panel.repaint();                
                return true;
            }
        }
        else
        if (action.equals("Clone"))
        {
        	VM3DObject o = locateNode(x,y);
            VMNode n = (o instanceof VMNode)? (VMNode)o: null; 
            if (n!=null)
            {
                VMNode nn=(VMNode)(n.newInstance(space.m_world));
                nn.m_x+=1;
                nn.m_y+=1;
                space.m_viewpoint.m_nodes.add(nn);
//System.out.println("Cloned!");
                reflectInsert(nn);//20060605
                panel.repaint();                
                return true;
            }
        }
        return false;
    }

}