(in-package :user) ;

(defun no-nil-string(x)
	(if x x "")
	)

(defun list-contains-string(string list)
	(dolist (list-string list)
		(if (equal list-string string)
			(return-from list-contains-string t)
			)
		)
	nil
	)

; composite key of integers from list in reverse order, first items low, last items top
(defun composite-key-of-list(int-list)
	(let ((key 0)(shift 0))
		(dolist (int-val int-list)
			(setq key (logior key (ash int-val shift)))
			(incf shift 32)
			)
		key
		);let
	)

; composite key of integers from list in straight order, first items top, last items low
(defun composite-key-of-list-straight(int-list)
	(let ((key 0))
		(dolist (int-val int-list)
			(setq key (logior (ash key 32) int-val))
			)
		key
		);let
	)

; composite key of two integer oid values given the order specified
(defun composite-key-of-two-oids(oid0 oid1)
	(logior (ash oid0 32) oid1)	; (+ (ASH oid0 32) oid1) will work with negative numbers too
	)

;(composite-key-of-two-oids 77 88)=>330712481880
;(decompose-key-to-two-oids 330712481880)=>77 88
(defun decompose-key-to-two-oids(key)
	(let ((oid0 (ash key -32)))
		(values oid0 (- key (ash oid0 32)))
		)
	)
	
; composite key of two integer oid values, less value first (left), greater value next (right)
(defun composite-key-of-two(arg0 arg1)
	(let ((oid0 (db-object-oid arg0))(oid1 (db-object-oid arg1))(oid))
		; make it so oid0 < oid1
		(if (> oid0 oid1)
			(let () (setq oid oid0)(setq oid0 oid1)(setq oid1 oid))
			)
		(logior (ash oid0 32) oid1)	; (+ (ASH oid0 32) oid1) will work with negative numbers too
		)
	)

; maybe need to rewrite using pointer to structure to avoid: 
; (incf (gethash "parrot" h 0)) consists of one gethash and one puthash
;
; John :  Before doing something like allocating an object in which to store a 
; value to be incremented, I'd first run the profiler to see if the two 
; second lookup in the incf is taking a significant amount of time.
(defun count-binary-evidence( table arg0 arg1 evidence )
	;(let ((key (composite-key-of-two-oids arg0 arg1)))
	;	(if (gethash key table)
	;		(incf (gethash key table) evidence)
	;		(setf (gethash key table) evidence)
	;		)
	;	)
	(incf (gethash (composite-key-of-two-oids arg0 arg1) table 0) evidence)
	)

; maybe need to rewrite using pointer to structure to avoid: 
; (incf (gethash "parrot" h 0)) consists of one gethash and one puthash
(defun binary-evidence( table arg0 arg1 )
	;(let ((e (gethash (composite-key-of-two-oids arg0 arg1) table)))
	;	(if e e 0) ; value or 0 if not found
	;	)
	(gethash (composite-key-of-two-oids arg0 arg1) table 0)
	)

;
;(extract-string "a" "abc" "c")=>"b"
;(extract-string "" "abc" "c")=>"ab"
;(extract-string nil "abc" "c")=>"ab"
;
(defun extract-string(prefix source suffix)
	(let (
		(start (search prefix source))
		(end (search suffix source))
		)
		(subseq source (+ start (length prefix)) end)
		)
	)

;http://www.ietf.org/rfc/rfc2396.txt
;public static String decode(String request) throws Exception
(defun decode-url(request)
	;StringBuffer sb = new StringBuffer();
	;int start=0;
	(let (
		(sb "")
		(start 0)
		)
		;for (;;)
		(do () ((= start -1));infinite?
			;int next = request.indexOf("%",start);
			(let ((next (position #\% request :start start)))
				;if (next==-1){
				;	sb.append(request.substring(start));
				;	break;}
				(if (not next)
					(return-from decode-url (concatenate 'string sb (subseq request start)))
				;else
					;if (next>0)
					;	sb.append(request.substring(start,next));
					;try {
					;	String code = request.substring(next+1,next+3); 
					;	sb.append((char)Integer.parseInt(code,16));
					;	start = next+3;
					;} catch (Exception e) {
					;	sb.append(request.substring(next,next+1));
					;	start = next+1;
					;}
					(let* (
						(code-string (subseq request (+ next 1) (+ next 3)))
						;(code-number (parse-integer code-string :radix 16 :junk-allowed t))
						(code-number)
						) 
						(setq code-number (parse-integer code-string :radix 16 :junk-allowed t))	
						(if (> next 0)
							(setq sb (concatenate 'string sb (subseq request start next)))
							)
						(if code-number
							(let ()
								(setq sb (concatenate 'string sb (string (code-char code-number))))
								(setq start (+ next 3))
								)
							(let ()
								(setq sb (concatenate 'string sb (subseq request next (1+ next))))
								(setq start (1+ next))
								)
							);if
						);let
					);if
				);let
			);loop
		sb
		);let
	)
	

; 
;(vector-of-lists-to-delimited-string (make-array 2 :element-type 'cons :initial-contents '((1 2) ("three" "four"))) "|" #\; )
;
(defun vector-of-lists-to-delimited-string(vector delim-outer delim-inner)
	(let ((res))
		(do ((i 0 (1+ i))) ((>= i (length vector))) 
			(setq res (concatenate 'string res (list-to-delimited-string (aref vector i) delim-inner) delim-outer))
			)
		res
		)
	)	

(defun vector-to-delimited-string(vector delim-outer)
	(let ((res))
		(do ((i 0 (1+ i))) ((>= i (length vector))) 
			(let ((str (string (aref vector i))))
				(setq res (concatenate 'string res str delim-outer))
				)
			)
		res
		)
	)	

	
(defun calc-proximity(a b)
	(if (and (numberp a) (numberp b))
		(let ((max (max (abs a) (abs b))))
			(if (= a b) 
				1
				(- 1 (float (/ (abs (- a b)) max)))
				)
			)
		(if (and (stringp a) (stringp b))
			(if (string-equal a b) 1 0)
			0
			)
		)
	)

;
; Average proximity of a list of numeric values:
; ( SUMi ( AVG - |Vi - AVG| ) ) / AVG
;
;(calc-list-proximity '(10 10 10 10)) => 1 
;(calc-list-proximity '( 5 10 15 10)) => 0.75
;(calc-list-proximity '( 4  8 12 16)) => 0.6 
;(calc-list-proximity '( 1 13 13 13)) => 0.55 
;(calc-list-proximity '( 5  5 15 15)) => 0.5
;(calc-list-proximity '( 1  6 14 19)) => 0.35
;(calc-list-proximity '( 1  1 19 19)) => 0.1 
;(calc-list-proximity '(0.1 0.1 0.1 39.7)) => -0.485
; 
(defun calc-list-proximity-old(list)
	(let ((sum 0)(cnt 0))
		(dolist (val list)
			(incf cnt)
			(incf sum val)
			)
		(if (= sum 0)
			nil
			(let ((avg (/ sum cnt))(sum-delta 0))
(print avg)
				(dolist (val list)
(print (list val (- val avg) (- avg (abs (- val avg)))))
					(incf sum-delta (- avg (abs (- val avg))))
					);dolist
				(float (/ sum-delta cnt avg))
				);let
			);if
		);let
	)

;
;(calc-list-proximity '(10 10 10 10)) => 1 
;(calc-list-proximity '( 5 10 15 10)) => 0.79
;(calc-list-proximity '( 4  8 12 16)) => 0.66 
;(calc-list-proximity '( 1 13 13 13)) => 0.6 
;(calc-list-proximity '( 5  5 15 15)) => 0.58
;(calc-list-proximity '( 1  6 14 19)) => 0.48
;(calc-list-proximity '( 1  1 19 19)) => 0.31 
;(calc-list-proximity '( 1  1  1 13))= 0.26
;(calc-list-proximity '(0.1 0.1 0.1 39.7)) => 0.07
;
(defun calc-list-proximity(list)
	(let ((sum 0)(cnt 0))
		(dolist (val list)
			(incf cnt)
			(incf sum val)
			)
		(if (= sum 0)
			nil
			(let ((avg (/ sum cnt)))
;(print avg)
				(setq sum 0)
				(dolist (val list)
;(print (list val (calc-proximity val avg)))
					(incf sum (calc-proximity val avg))
					);dolist
				(float (/ sum cnt))
				);let
			);if
		);let
	)

;
; uniformity = SUM( Ei * Ei ) / SUM(Ei) * SUM(Ei)
;
;(setq x (make-hash-table)) (setf (gethash 1 x) 12) (setf (gethash 2 x) 0) (calc-table-uniformity x)
;=>1
;(setq x (make-hash-table)) (setf (gethash 1 x) 10) (setf (gethash 2 x) 2) (calc-table-uniformity x)
;=>0.72
;(setq x (make-hash-table)) (setf (gethash 1 x)  8) (setf (gethash 2 x) 4) (calc-table-uniformity x)
;=>0.55
;(setq x (make-hash-table)) (setf (gethash 1 x)  6) (setf (gethash 2 x) 6) (calc-table-uniformity x)
;=>0.5
;
(defun calc-table-uniformity(table)
	(let ((e 0)(ee 0))
		(maphash #'(lambda(uom-id evidence) 
			(if uom-id (let () ; fool compiler so it does not cr
				(incf ee (* evidence evidence)) 
				(incf e evidence)
				))
			) table)
;(print (list ee (* e e)))
		(if (= e 0)
			nil
			(float (/ ee (* e e)))
			)
		)
	)


(defun list-avgerages-init(sum-vec cnt-vec)	
	(do ((i 0 (1+ i))) ((>= i (length sum-vec))) 
		(setf (aref sum-vec i) 0)
		(setf (aref cnt-vec i) 0)
		)
	)

(defun list-avgerages-incf(sum-vec cnt-vec val-list)
	(let ((i 0))
		(dolist (val val-list)
			(if (and val (numberp val)) (let ()
				(incf (aref sum-vec i) val)	
				(incf (aref cnt-vec i))	
				))
			(incf i)
			)
		)
	)

;
;(defvar sum (make-array 3 :element-type 'real))
;(defvar cnt (make-array 3 :element-type 'real))
;(list-avgerages-init sum cnt)
;(list-avgerages-incf sum cnt '(1 nil 10))
;(list-avgerages-incf sum cnt '(2 nil nil))
;(list-avgerages-incf sum cnt '(3 nil 30))
;(list-avgerages-list sum cnt)
;
(defun list-avgerages-list(sum-vec cnt-vec)	
	(let ((avg-list))
		(do ((i 0 (1+ i))) ((>= i (length sum-vec))) 
			(push 
				(if (> (aref cnt-vec i) 0)
					(float (/ (aref sum-vec i) (aref cnt-vec i)))
					nil
					)
				avg-list)
			)
		(reverse avg-list)
		)
	)

;
;(list-nil-to-empty-string '(nil 1 "two" nil 3 "four" nil))
;
(defun list-nil-to-empty-string(lst)
	(let ((l lst))
		(do () ((not l))
			(if (not (first l))
				(rplaca l "")
				)	
			(setq l (rest l))
			)
		lst
		)
	) 
	
;
;checks if child contains all items of parent assuming both are ordered in the same way
;(super '(1 2 3) '(4 1 5 2 7 3 8)) => t
;(super '(1 2 3) '(4 1 5 7 3 8)) => t
;
(defun super(parent child)
	(dolist (c child)
;(print (list c parent))
		(if (equal c (first parent)) ; if match
			(setq parent (rest parent)) ; advance parent
			)
		)
	(not parent)	
	)
	
;
;(reverse-list-from-vector '(1 3 5) (make-array 7 :element-type 'string :initial-contents '("A" "B" "C" "D" "E" "F" "G")))
;	
(defun reverse-list-from-vector(aref-list source-vector)	
	(let ((retlist))
		(dolist (i aref-list)
			(let ((id (aref source-vector i)))
				(push (if id id 0) retlist)
				)
			)
		retlist
		)
	)

;
;(decimal-granularity 300)=>10.0
;(decimal-granularity 600)=>100.0
;	
(defun decimal-granularity(max)
	(float (/ (expt 10 (round (log max 10))) 10))
	)
	
;
;(granular-value 10 123)=>120.0
;(granular-value 100 123)=>100.0
;
(defun granular-value(granularity value)
	(float (* (round (/ value granularity)) granularity))
	)

;
;(valid-file-name "V|M|N")
;(valid-file-name "V=V1|M=M1|N=N1")
;
(defun valid-file-name(str)
	(substitute #\- #\| (substitute #\_ #\= str))
	)
	
;
;(format-assoc-list '((A . B)("X" . "Y")(1 . 2)("3" . "4")))
;
(defun format-assoc-list(list)
	(let (
		(s (make-string-output-stream))
		(started)
		)
		(dolist (pair list) 
			(if started (format s "|"))
			(setq started t)
			(format s "~A=~A" (first pair) (rest pair))
			)
		(get-output-stream-string s)
		)
	)
	
